'
' Crownhill Proton compiler 8-bit fixed point FFT routine
' For use with version 3.5.3.0 onwards of the compiler and 18F devices.
'
' Written by Les Johnson 02-02-2012
'
' You are free to use this code as you see fit, however, a mention of the author and compiler would be appreciated.
' The code is offered as-is and Crownhill nor the author take any responsibility for its use.
'
$ifndef FFT_Elements
    $define FFT_Elements 64                         ' Default elements if no define is used in the main program
$endif
'
' Trap the element value
'
$if (FFT_Elements <> 128) And (FFT_Elements <> 64) And (FFT_Elements <> 32) And (FFT_Elements <> 16)
    $error "Incorret amount of elements. 128, 64, 32 or 16 required"
$endif

    $define FFT_NWave $eval FFT_Elements * 2        ' Calculate the size of the Sin and Cos waves required
'
' Create variables used for the merging of imaginary and real data
'
    Dim FFT_wReal As Word System
    Dim FFT_wImaginary As Word System
    Dim FFT_bImaginary As FFT_wImaginary.Byte0
    Dim FFT_bReal As FFT_wReal.Byte0
'
' Create variables used by Fix_FFT
' Made mostly system types so they stay is bankless Access RAM for speed
'
    Dim FFT_bTemp1 As Byte System
    Dim FFT_bTemp2 As Byte System
    Dim FFT_bTemp3 As Byte System
    Dim FFT_bTemp4 As Byte System

    Dim FFT_bM As Byte System
    Dim FFT_bMR As Byte System
    Dim FFT_bJ As Byte System
    Dim FFT_bL As Byte System
    Dim FFT_bK As Byte System

    Dim FFT_bIstep As Byte System
    Dim FFT_bIndex As Byte System

    Dim FFT_bQR As Byte System
    Dim FFT_bQI As Byte System
    Dim FFT_bTempReal As Byte System
    Dim FFT_bTempImaginary As Byte System
    Dim FFT_bWR As Byte System
    Dim FFT_bWI As Byte System

    Dim FFT_bRealData[FFT_Elements] As Byte Heap        ' Holds the real data
    Dim FFT_bImaginaryData[FFT_Elements] As Byte Heap   ' Holds the imaginary data

    Dim FXPM_wPROD As PRODL.Word                        ' Make a 16-bit variable from PRODL\H
   
    Symbol cFFT_NumberOfSamples = FFT_Elements          ' The amount of elements in the FFT
    Symbol cFFT_NumberOfRealElements = FFT_Elements / 2 ' The amount of real elements after the FFT has performed its task 
    Symbol cNWave_Div4 = FFT_NWave / 4                  ' Length of SineWave divided by 4
     
$if FFT_NWave = 256
    $define cLog2_Nwave 8                               ' Log2(256)
$elseif FFT_NWave = 128
    $define cLog2_Nwave 7                               ' Log2(128)
$elseif FFT_NWave = 64
    $define cLog2_Nwave 6                               ' Log2(64)
$elseif FFT_NWave = 32
    $define cLog2_Nwave 5                               ' Log2(32)
$elseif FFT_NWave = 16
    $define cLog2_Nwave 4                               ' Log2(16)
$endif
'--------------------------------------------------------------------------------
    GoTo _FFT_Main_                                     ' Jump over the subroutines
'--------------------------------------------------------------------------------
' 8-bit Signed Fixed-Point multiplication & scaling.
' Input     : pMultiplier holds the 8-bit value to multiply
'           : pMultiplicand holds the 8-Bit value to multiply with
' Output    ' pResult holds the resulting product of the multiplication
' Notes     : Scaling ensures that result remains within 16-bits.
'           : The code is rolled out for speed instead of using loops
'
$define Fixed_Multiply(pMultiplier, pMultiplicand, pResult) '
    If pMultiplicand = 0 Then       '
        pResult = 0                 '
    ElseIf pMultiplier = 0 Then     '
        pResult = 0                 '
    Else                            '
        Movf pMultiplier,w,0        '
        Mulwf pMultiplicand,0       '
        Btfsc pMultiplicand,7,0     '
        Subwf PRODH,f,0             '
        Btfss pMultiplier,7,0       '
        Bra $ + 6                   '
        Movf pMultiplicand,w,0      '
        Subwf  PRODH,f,0            '
        Rrcf   PRODH,f,0            '
        Rrcf   PRODL,f,0            '
        Rrcf   PRODH,f,0            '
        Rrcf   PRODL,f,0            '
        Rrcf   PRODH,f,0            '
        Rrcf   PRODL,f,0            '
        Rrcf   PRODH,f,0            '
        Rrcf   PRODL,f,0            '
        Rrcf   PRODH,f,0            '
        Rrcf   PRODL,f,0            '
        Rrcf   PRODH,f,0            '
        Rrcf   PRODL,f,0            '
        Movlw  3                    '
        Andwf  PRODH,f,0            '
        WREG = PRODL & 1            '
        Bcf    STATUS,C,0           '
        Rrcf   PRODH,f,0            '
        Rrcf   PRODL,f,0            '
        Btfsc  PRODH,6,0            '
        Bsf    PRODH,7,0            '
        pResult = FXPM_wPROD + WREG '
    EndIf

'-------------------------------------------------------------------------
' Create 8-bit sine wave data
'
SineWave:
$if FFT_NWave = 256
    CData As Byte 000, -002, -003, -005, -006, -008, -009, -011,_
                  -012, -014, -016, -017, -019, -020, -022, -023,_
                  -024, -026, -027, -029, -030, -032, -033, -034,_
                  -036, -037, -038, -039, -041, -042, -043, -044,_
                  -045, -046, -047, -048, -049, -050, -051, -052,_
                  -053, -054, -055, -056, -056, -057, -058, -059,_
                  -059, -060, -060, -061, -061, -062, -062, -062,_
                  -063, -063, -063, -064, -064, -064, -064, -064,_
                  -064, -064, -064, -064, -064, -064, -063, -063,_
                  -063, -062, -062, -062, -061, -061, -060, -060,_
                  -059, -059, -058, -057, -056, -056, -055, -054,_
                  -053, -052, -051, -050, -049, -048, -047, -046,_
                  -045, -044, -043, -042, -041, -039, -038, -037,_
                  -036, -034, -033, -032, -030, -029, -027, -026,_
                  -024, -023, -022, -020, -019, -017, -016, -014,_
                  -012, -011, -009, -008, -006, -005, -003, -002,_
                  000, 002, 003, 005, 006, 008, 009, 011,_
                  012, 014, 016, 017, 019, 020, 022, 023,_
                  024, 026, 027, 029, 030, 032, 033, 034,_
                  036, 037, 038, 039, 041, 042, 043, 044,_
                  045, 046, 047, 048, 049, 050, 051, 052,_
                  053, 054, 055, 056, 056, 057, 058, 059,_
                  059, 060, 060, 061, 061, 062, 062, 062,_
                  063, 063, 063, 064, 064, 064, 064, 064,_
                  064, 064, 064, 064, 064, 064, 063, 063,_
                  063, 062, 062, 062, 061, 061, 060, 060,_
                  059, 059, 058, 057, 056, 056, 055, 054,_
                  053, 052, 051, 050, 049, 048, 047, 046,_
                  045, 044, 043, 042, 041, 039, 038, 037,_
                  036, 034, 033, 032, 030, 029, 027, 026,_
                  025, 023, 022, 020, 019, 017, 016, 014,_
                  012, 011, 009, 008, 006, 005, 003, 002
$elseif FFT_NWave = 128
    CData As Byte 000, -003, -006, -009, -012, -016, -019, -022,_
                  -024, -027, -030, -033, -036, -038, -041, -043,_
                  -045, -047, -049, -051, -053, -055, -056, -058,_
                  -059, -060, -061, -062, -063, -063, -064, -064,_
                  -064, -064, -064, -063, -063, -062, -061, -060,_
                  -059, -058, -056, -055, -053, -051, -049, -047,_
                  -045, -043, -041, -038, -036, -033, -030, -027,_
                  -024, -022, -019, -016, -012, -009, -006, -003,_
                  000, 003, 006, 009, 012, 016, 019, 022,_
                  024, 027, 030, 033, 036, 038, 041, 043,_
                  045, 047, 049, 051, 053, 055, 056, 058,_
                  059, 060, 061, 062, 063, 063, 064, 064,_
                  064, 064, 064, 063, 063, 062, 061, 060,_
                  059, 058, 056, 055, 053, 051, 049, 047,_
                  045, 043, 041, 038, 036, 033, 030, 027,_
                  025, 022, 019, 016, 012, 009, 006, 003
$elseif FFT_NWave = 64
    CData As Byte 000, -006, -012, -019, -024, -030, -036, -041,_
                  -045, -049, -053, -056, -059, -061, -063, -064,_
                  -064, -064, -063, -061, -059, -056, -053, -049,_
                  -045, -041, -036, -030, -024, -019, -012, -006,_
                  000, 006, 012, 019, 024, 030, 036, 041,_
                  045, 049, 053, 056, 059, 061, 063, 064,_
                  064, 064, 063, 061, 059, 056, 053, 049,_
                  045, 041, 036, 030, 025, 019, 012, 006
$elseif FFT_NWave = 32
    CData As Byte 000, -012, -024, -036, -045, -053, -059, -063,_
                  -064, -063, -059, -053, -045, -036, -024, -012,_
                  000, 012, 024, 036, 045, 053, 059, 063,_
                  064, 063, 059, 053, 045, 036, 025, 012
$elseif FFT_NWave = 16
    CData As Byte 000, -024, -045, -059, -064, -059, -045, -024,_
                  000, 024, 045, 059, 064, 059, 045, 025
$else
    $error "Unknown FFT_NWave length"
$endif
'-------------------------------------------------------------------------
' Create 8-bit cosine wave data
'
CosineWave:
$if FFT_NWave = 256
    CData As Byte 064, 064, 064, 064, 064, 064, 063, 063,_
                  063, 062, 062, 062, 061, 061, 060, 060,_
                  059, 059, 058, 057, 056, 056, 055, 054,_
                  053, 052, 051, 050, 049, 048, 047, 046,_
                  045, 044, 043, 042, 041, 039, 038, 037,_
                  036, 034, 033, 032, 030, 029, 027, 026,_
                  024, 023, 022, 020, 019, 017, 016, 014,_
                  012, 011, 009, 008, 006, 005, 003, 002,_
                  000, -002, -003, -005, -006, -008, -009, -011,_
                  -012, -014, -016, -017, -019, -020, -022, -023,_
                  -024, -026, -027, -029, -030, -032, -033, -034,_
                  -036, -037, -038, -039, -041, -042, -043, -044,_
                  -045, -046, -047, -048, -049, -050, -051, -052,_
                  -053, -054, -055, -056, -056, -057, -058, -059,_
                  -059, -060, -060, -061, -061, -062, -062, -062,_
                  -063, -063, -063, -064, -064, -064, -064, -064,_
                  -064, -064, -064, -064, -064, -064, -063, -063,_
                  -063, -062, -062, -062, -061, -061, -060, -060,_
                  -059, -059, -058, -057, -056, -056, -055, -054,_
                  -053, -052, -051, -050, -049, -048, -047, -046,_
                  -045, -044, -043, -042, -041, -039, -038, -037,_
                  -036, -034, -033, -032, -030, -029, -027, -026,_
                  -024, -023, -022, -020, -019, -017, -016, -014,_
                  -012, -011, -009, -008, -006, -005, -003, -002,_
                  000, 002, 003, 005, 006, 008, 009, 011,_
                  012, 014, 016, 017, 019, 020, 022, 023,_
                  024, 026, 027, 029, 030, 032, 033, 034,_
                  036, 037, 038, 039, 041, 042, 043, 044,_
                  045, 046, 047, 048, 049, 050, 051, 052,_
                  053, 054, 055, 056, 056, 057, 058, 059,_
                  059, 060, 060, 061, 061, 062, 062, 062,_
                  063, 063, 063, 064, 064, 064, 064, 064
$elseif FFT_NWave = 128
    CData As Byte 064, 064, 064, 063, 063, 062, 061, 060,_
                  059, 058, 056, 055, 053, 051, 049, 047,_
                  045, 043, 041, 038, 036, 033, 030, 027,_
                  024, 022, 019, 016, 012, 009, 006, 003,_
                  000, -003, -006, -009, -012, -016, -019, -022,_
                  -024, -027, -030, -033, -036, -038, -041, -043,_
                  -045, -047, -049, -051, -053, -055, -056, -058,_
                  -059, -060, -061, -062, -063, -063, -064, -064,_
                  -064, -064, -064, -063, -063, -062, -061, -060,_
                  -059, -058, -056, -055, -053, -051, -049, -047,_
                  -045, -043, -041, -038, -036, -033, -030, -027,_
                  -024, -022, -019, -016, -012, -009, -006, -003,_
                  000, 003, 006, 009, 012, 016, 019, 022,_
                  024, 027, 030, 033, 036, 038, 041, 043,_
                  045, 047, 049, 051, 053, 055, 056, 058,_
                  059, 060, 061, 062, 063, 063, 064, 064
$elseif FFT_NWave = 64
    CData As Byte 064, 064, 063, 061, 059, 056, 053, 049,_
                  045, 041, 036, 030, 024, 019, 012, 006,_
                  000, -006, -012, -019, -024, -030, -036, -041,_
                  -045, -049, -053, -056, -059, -061, -063, -064,_
                  -064, -064, -063, -061, -059, -056, -053, -049,_
                  -045, -041, -036, -030, -024, -019, -012, -006,_
                  000, 006, 012, 019, 024, 030, 036, 041,_
                  045, 049, 053, 056, 059, 061, 063, 064
$elseif FFT_NWave = 32
    CData As Byte 064, 063, 059, 053, 045, 036, 024, 012,_
                  000, -012, -024, -036, -045, -053, -059, -063,_
                  -064, -063, -059, -053, -045, -036, -024, -012,_
                  000, 012, 024, 036, 045, 053, 059, 063
$elseif FFT_NWave = 16
    CData As Byte 064, 059, 045, 024, 000, -024, -045, -059,_
                  -064, -059, -045, -024, 000, 024, 045, 059
$else
    $error "Unknown FFT_NWave length"
$endif

'-------------------------------------------------------------------------
' Perform a forward Fast Fourier Transform
' Input     : Array FFT_bRealData holds the 8-bit ADC samples to perform the FFT upon
' Output    : Array FFT_bRealData holds real data from the FFT
'           : Array FFT_bImaginaryData holds imaginary data from the FFT
' Notes     : None
'
$define Fix_FFT() GoSub _Fix_FFT

_Fix_FFT:
    Clear FFT_bImaginaryData                    ' Clear the imaginary data array
'
' Decimation in time - re-order data
'
    FFT_bMR = 0
    FFT_bM = 1
    Repeat
        FFT_bL = cFFT_NumberOfSamples
        Repeat
            FFT_bL = FFT_bL >> 1
            FFT_bTemp1 = FFT_bMR + FFT_bL
        Until FFT_bTemp1 < cFFT_NumberOfSamples
        FFT_bTemp1 = FFT_bL - 1
        FFT_bMR = FFT_bMR & FFT_bTemp1
        FFT_bMR = FFT_bMR + FFT_bL
        If FFT_bMR > FFT_bM Then
            FFT_bTempReal = FFT_bRealData[FFT_bM]
            FFT_bRealData[FFT_bM] = FFT_bRealData[FFT_bMR]
            FFT_bRealData[FFT_bMR] = FFT_bTempReal
        EndIf
        Inc FFT_bM
    Until FFT_bM >= cFFT_NumberOfSamples
'
' Perform the FFT
'
    FFT_bL = 1
    FFT_bK = cLog2_Nwave - 1
    While FFT_bL < cFFT_NumberOfSamples
        FFT_bIstep = FFT_bL << 1
        FFT_bM = 0
        Repeat
            FFT_bJ = FFT_bM << FFT_bK
            FFT_bWR = CRead8 CosineWave[FFT_bJ]                     ' Cosine
            FFT_bWI = CRead8 SineWave[FFT_bJ]                       ' Sine
            FFT_bIndex = FFT_bM
            Repeat
                FFT_bJ = FFT_bIndex + FFT_bL

                FFT_bTemp3 = FFT_bRealData[FFT_bJ]
                FFT_bTemp4 = FFT_bImaginaryData[FFT_bJ]

                Fixed_Multiply(FFT_bWR, FFT_bTemp3 , FFT_bTemp1)    ' 8-bit Fixed point multiply into FFT_bTemp1
                Fixed_Multiply(FFT_bWI, FFT_bTemp4 , FFT_bTemp2)    ' 8-bit Fixed point multiply  into FFT_bTemp2
                FFT_bTempReal = FFT_bTemp1 - FFT_bTemp2

                Fixed_Multiply(FFT_bWR, FFT_bTemp4 , FFT_bTemp1)    ' 8-bit Fixed point multiply into FFT_bTemp1
                Fixed_Multiply(FFT_bWI, FFT_bTemp3 , FFT_bTemp2)    ' 8-bit Fixed point multiply into FFT_bTemp2
                FFT_bTempImaginary = FFT_bTemp1 + FFT_bTemp2

                FFT_bQR = FFT_bRealData[FFT_bIndex]
                FFT_bQR = FFT_bQR.SByte / 2                         ' Reduce the chance of overflow
                FFT_bQI = FFT_bImaginaryData[FFT_bIndex]
                FFT_bQI = FFT_bQI.SByte / 2                         ' Reduce the chance of overflow
                FFT_bRealData[FFT_bJ] = FFT_bQR - FFT_bTempReal
                FFT_bImaginaryData[FFT_bJ] = FFT_bQI - FFT_bTempImaginary
                FFT_bRealData[FFT_bIndex] = FFT_bQR + FFT_bTempReal
                FFT_bImaginaryData[FFT_bIndex] = FFT_bQI + FFT_bTempImaginary
                FFT_bIndex = FFT_bIndex + FFT_bIstep
            Until FFT_bIndex >= cFFT_NumberOfSamples
            Inc FFT_bM
        Until FFT_bM >= FFT_bL
        FFT_bK = FFT_bK - 1
        FFT_bL = FFT_bIstep
    Wend
' Fall through to "_ConvertToReal"
'-------------------------------------------------------------
' Convert the real and imaginary arrays to a single real array
' Input     : Array FFT_bRealData holds the real data from the FFT
'           : Array FFT_bImaginaryData holds the imaginary data from the FFT
' Output    : Results are placed back into the 'FFT_bRealData' array
' Notes     : There are two ways to compute the real part:
'           : 1). Find the real value by the square root of the squares of real and imaginary values
'           : 2). Find the real value by summing the real and imaginary parts (quicker, but dirty)
'           : Also removed the first element which is the DC component
'
_ConvertToReal:
    FFT_bIndex = 0
    Repeat
        FFT_bReal = FFT_bRealData[FFT_bIndex]                       ' Extract a value from the real array
        FFT_bReal = Abs FFT_bReal                                   ' Make sure it's positive
        FFT_bImaginary = FFT_bImaginaryData[FFT_bIndex]             ' Extract a value from the imaginary array
        FFT_bImaginary = Abs FFT_bImaginary                         ' Make sure it's positive
        '
        ' Quick and dirty method
        '
$ifndef FFT_Use_Sqr
        FFT_bRealData[FFT_bIndex] = FFT_bReal + FFT_bImaginary      ' Place the summed value back into the RealData array (quick and dirty method)
$else
        '
        ' Slower but more refined method (also traps zero values for extra speed)
        '
        If FFT_bImaginary = 0 Then                                  ' Is FFT_bImaginary zero?
            If FFT_bReal = 0 Then                                   ' Yes. So is FFT_bReal zero?
                FFT_bRealData[FFT_bIndex] = 0                       ' Yes. So place a zero into the array
            Else                                                    ' Otherwise...
                FFT_wReal = FFT_bReal * FFT_bReal                   ' Square the real value
                FFT_bRealData[FFT_bIndex] = ISqr FFT_wReal          ' Find the square root of real
            EndIf
        ElseIf FFT_bReal = 0 Then                                   ' Otherwise. Is FFT_bReal zero?
            If FFT_bImaginary = 0 Then                              ' Yes. So is FFT_bImaginary zero?
                FFT_bRealData[FFT_bIndex] = 0                       ' Yes. So place a zero into the array
            Else                                                    ' Otherwise...
                FFT_wImaginary = FFT_bImaginary * FFT_bImaginary    ' Square the imaginary value
                FFT_bRealData[FFT_bIndex] = ISqr FFT_wImaginary     ' Find the square root of imaginary
            EndIf
        Else                                                        ' Otherwise. None are zero
            FFT_wImaginary = FFT_bImaginary * FFT_bImaginary        ' Square the imaginary value
            FFT_wReal = FFT_bReal * FFT_bReal                       ' Square the real value
            FFT_bRealData[FFT_bIndex] = ISqr FFT_wImaginary + FFT_wReal ' Find the square root of real and imaginary
        EndIf
$endif
        Inc FFT_bIndex
    Until FFT_bIndex >= cNWave_Div4
    FFT_bRealData#0 = 0                                             ' Remove the DC content from the first element
    Return
'-------------------------------------------------------------------------
_FFT_Main_:


